; Boards/ARMulator/DefinesH.s
; Defines various constants and macros for the ARMulator
;
; 19/Feb/1999 ned

; KEEP CONSISTENT WITH BOARDDEFINES.h


; Number of handlers for each exception
UndefinedExcptHandlers	EQU	2
SWIExcptHandlers	EQU	2
PAbortExcptHandlers	EQU	2
DAbortExcptHandlers	EQU	2
IRQExcptHandlers	EQU	2
FIQExcptHandlers	EQU	2

; Constants - Exception related
ResetV		EQU	0
UndefV		EQU	1
SWIV		EQU	2
PAbortV		EQU	3
DAbortV		EQU	4
IRQV		EQU	6
FIQV		EQU	7

; Sizes of the processor mode stacks
IRQ32StackSize		EQU	4096
FIQ32StackSize		EQU	1024
Abort32StackSize	EQU	1024
Undef32StackSize	EQU	1024
SVC32StackSize		EQU	8192
BOARD_TOTALSTKSSIZE	EQU	IRQ32StackSize + FIQ32StackSize + Abort32StackSize + Undef32StackSize + SVC32StackSize

; Location of exception handlers
BOARD_EXCEPTHLOCATION		EQU	0x0
; Location of HAL r/w data
BOARD_HALZPLOCATION		EQU	0xc0
; Location of UART ctrl blks
BOARD_HALZP_UARTSPACE		EQU	0x1000				; 4k
; Location of L1 MMU page tables
BOARD_ARMMMU_L1PAGETABLES	EQU	0x4000				; 16k (16k of L1 page tables)
; Where we put the processor stacks
BOARD_HALSTKSLOCATION		EQU	0x8000				; 32k (16k)

; Locations of UART zero page locations
HALZP_ANGELUARTSPACE		EQU	(0 * HALUARTBLK_SIZEOF)
HALZP_DEMONUARTSPACE		EQU	(1 * HALUARTBLK_SIZEOF)

; This macro is called to remap ROM to its proper place and offset the PC if necessary
	MACRO
	BOARD_MOVEROM

	MEND

; This macro is called to initialise DRAM for use (enable power to it, whatever)
	MACRO
	BOARD_INITIALISEDRAM

	MEND

; This macro is called to return the location of the zero page memory
; NOTE: This gets called in the general ISR so make it quick!
; It returns a pointer to 64k of zero page memory in R0
	MACRO
	BOARD_RETURNZPLOCATION $reg
	LDR	$reg,=(0x00080000 - 0x10000)		; 64k before mem end

	MEND

; This macro is called to set up the memory system of the board
; This may entail: (i) determining memory configuration (ii) building page tables appropriately
; HOWEVER: if no MMU is present, or not desired, then the processor caches may not be enabled
	MACRO
	BOARD_SETUPMEMORYSYSTEM

; --- Determine memory configuration
	; This board has a fixed memory config (ARMulator, 2Mb)

; --- Build page tables
	[ 1=0
	BOARD_RETURNZPLOCATION R9
	ADD	R0,R9,#HALZP_ARMMMU_L1PAGETABLES
	BL	ARMMMU_BeginL1PageTableBuild

	; Map 0x00000000 to 0x00200000 to same (ROM)
	MOV	R1,#0				; start Mb no
	MOV	R2,#0				; start desired virtual addr
	MOV	R3,#0x00200000			; end desired virtual addr
	MOV	R4,#0xa				; domain, access permissions, caching & buffering of section
	BL	ARMMMU_MakeL1PageTableEntries
	
	; Map 0x00200000 to 0x08000000 to Fault
	MOV	R3,#0x08000000			; end desired virtual addr
	MOV	R4,#0				; domain of section
	BL	ARMMMU_MakeL1PageTableEntries
	
	; Map 0x08000000 to 0x08200000 to 0xc0000000 (DRAM)
	MOV	R2,#0x08000000			; start desired virtual addr
	MOV	R3,#0x08200000			; end desired virtual addr
	MOV	R4,#0				; domain of section
	BL	ARMMMU_MakeL1PageTableEntries

	; Map 0x08200000 to 0x00000000 to Fault
	MOV	R3,#0x00000000			; end desired virtual addr
	MOV	R4,#0				; domain of section
	BL	ARMMMU_MakeL1PageTableEntries

	; Set the MMU to use these tables
	MOV	R2,R0
	LDR	R0,=ARMMMU_SETTTB :OR: ARMMMU_SETMMU :OR: ARMMMU_SETWB :OR: ARMMMU_SETROM
	LDR	R1,=ARMMMU_SETTTB :OR: ARMMMU_SETMMU :OR: ARMMMU_SETWB :OR: ARMMMU_SETROM
	BL	ARMMMU_Set
	]

; --- Enable caches
	LDR	R0,=ARMCACHE_SETICACHE :OR: ARMCACHE_SETDCACHE
	LDR	R1,=ARMCACHE_SETICACHE :OR: ARMCACHE_SETDCACHE
	BL	ARMCache_Set

; --- Set location and extent of main memory when MMU is enabled
	MOV	R0,#0x00000000				; Base of memory on ARMulator is 0x0 and it extends 2Mb
	STR	R0,[R9,#HALZP_DRAMBASE]
	MOV	R0,#0x00200000
	STR	R0,[R9,#HALZP_DRAMTOP]

	MEND

; This macro is called to disable an unrecognised interrupt
; You have R0-R5 to play with
	MACRO
	BOARD_DISABLEUNKNOWNINTERRUPT

	MEND

; This macro is called to do any board-specific initialisations
; It is called just prior to calling NedHAL_Entry
; Keep within APCS regulations
	MACRO
	BOARD_INITIALISE

	MEND

; This macro is called to do any board-specific shutdowns
; It is called when NedHAL_Entry returns
; Keep within APCS regulations
	MACRO
	BOARD_FINALISE

	MEND


	END
