; Boards/AT91/Timer.s
; Timer module for the AT91 board
;
; 29/Dec/1999 ned

	AREA	AT91Timer, CODE, READONLY

	; Include zero page map
	INCLUDE Core/ARM/CommonH.s

	; Exports from this module
	EXPORT	AT91Timer_Initialise
	EXPORT	AT91Timer_AddHandler
	EXPORT	AT91Timer_Finalise

AT91Timer_Initialise
	LDR	R3,=0xfffe0000				; addr of Timer0 ctrl blk

	; The timers provided by the AT91 are very flexible
	; We just want an interrupt generated every 1cs
	MOV	R0,#0
	LDR	R1,=AT91Timer_HandlerAddr
	STR	R0,[R1]

	; Set Timer0 to Waveform mode using MCLK/1024 and
	; auto-reset on RC compare true
	LDR	R0,=(1<<15)+(1<<14)+4
	STR	R0,[R3,#0x04]				; TC_CMR
	; Set RC compare reg (the latch) to 100Hz (=320 @ MCLK/1024)
	LDR	R0,=320
	STR	R0,[R3,#0x1c]				; TC_RC

	; Install Timer0 interrupt handler (=IRQ4)
	; Use the AT91 AIC as it makes life so much easier
	LDR	R2,=0xfffff000				; base of AIC
	LDR	R0,=AT91Timer_ISR
	STR	R0,[R2,#0x90]				; AIC vector addr
	LDR	R0,=(1<<5) + 1
	STR	R0,[R2,#0x10]				; AIC int type

	; Enable interrupts
	LDR	R0,=(1<<4)
	STR	R0,[R2,#0x120]				; AIC enable int no 4
	; Enable RC compare true interrupt
	STR	R0,[R3,#0x24]				; TC_IER

	; Finally, enable timer0
	MOV	R0,#1					; enable command
	STR	R0,[R3,#0x0]				; TC_CCR
	MOV	R0,#4					; trigger command
	STR	R0,[R3,#0x0]				; TC_CCR
	MOV	PC, R14

AT91Timer_Finalise
	LDR	R3,=0xfffe0000
	LDR	R2,=0xfffff000
	; Disable Timer0
	MOV	R0,#2					; disable command
	STR	R0,[R3,#0x0]				; TC_CCR
	; Disable interrupts (IRQ4)
	LDR	R0,=(1<<4)
	STR	R0,[R3,#0x28]				; TC_IDR
	STR	R0,[R2,#0x124]				; AIC disable int no 4
	MOV	PC, R14

AT91Timer_AddHandler
	; On entry R0=handler addr, R1=value to pass, R2=pos
	LDR	R1,=AT91Timer_HandlerAddr
	STR	R0,[R1]
	MOV	PC,R14

AT91Timer_ISR
	; Thanks to the AT91 AIC, we are guaranteed only to be called when
	; Timer0 has triggered
	STMFD	R13!,{R0-R12,R14}
	MRS	R0,SPSR
	STMFD	R13!,{R0}
	LDR	R3,=0xfffe0000

	; Service the interrupt
	LDR	R0,[R3,#0x20]				; TC_SR

	[ :DEF: ANGEL
	; If Angel is present, some trickery is required to prevent
	; everything dying
	IMPORT	|Image$$RO$$Base|
	IMPORT	|Image$$RO$$Limit|

	LDR	R1,=|Image$$RO$$Base|
	LDR	R2,=|Image$$RO$$Limit|
	CMP	R14,R1
	BLO	AT91Timer_ISR_exit
	CMP	R14,R2
	BHS	AT91Timer_ISR_exit
	]

	; Code to do timer stuff here (can reenable interrupts maybe?)
	LDR	R1,=AT91Timer_HandlerAddr
	LDR	R1,[R1]
	CMP	R1,#0
	MOVNE	R0,#0
	MOVNE	R14,PC
	MOVNE	PC,R1
AT91Timer_ISR_exit
	; Tell the AIC we're done (must have interrupts disabled again)
	LDR	R2,=0xfffff000
	STR	R0,[R2,#0x130]				; AIC_EOICR
	LDMFD	R13!,{R0}
	MSR	SPSR,R0
	LDMFD	R13!,{R0-R12,R14}
	SUBS	PC,R14,#4

	AREA	AT91TimerData, DATA, READWRITE
AT91Timer_HandlerAddr
	DCD	0

	END
