; Core/ARM/Entry.s
; Initial entry point of CPU on reset
;
; 25/Sep/1999 ned

	; Include Zero Page Layout
	INCLUDE Core/ARM/CommonH.s

	; Include defines for selected ARM
	INCLUDE	ARMCoreH.s
	INCLUDE ARMMMUH.s
	INCLUDE ARMCacheH.s
	INCLUDE	Core/io/DebugIOH.s

	; Import used routines
	IMPORT	ARMCore_Initialise
	IMPORT	ARMCore_Set
	IMPORT	ARMMMU_Initialise
	IMPORT	ARMMMU_Set
	IMPORT	ARMMMU_BeginL1PageTableBuild
	IMPORT	ARMMMU_MakeL1PageTableEntries
	IMPORT	ARMCache_Initialise
	IMPORT	ARMCache_Set
	IMPORT	DbgIO_Initialise
	IMPORT	DbgIO_Finalise
	IMPORT	DbgIO_Set
	IMPORT	DbgIO_Write0
	IMPORT	DbgIO_WriteF
	IMPORT	ExceptH_Initialise
	IMPORT	ExceptH_AddHandler
	IMPORT	ExceptH_Finalise
	IMPORT	ExceptH_EntryUndefV
	IMPORT	ExceptH_EntrySWIV
	IMPORT	ExceptH_EntryPAbortV
	IMPORT	ExceptH_EntryDAbortV
	IMPORT	ExceptH_EntryIRQV
	IMPORT	ExceptH_EntryFIQV


	AREA	|!!!Entry|, CODE, READONLY

	EXPORT	|_main|
	EXPORT	|__main|
	EXPORT	NedHAL_EmergencyStop

	IMPORT	NedHAL_Entry
	IMPORT	|Image$$RO$$Base|
	IMPORT	|Image$$RO$$Limit|
	IMPORT	|Image$$RW$$Base|
	IMPORT	|Image$$RW$$Limit|
	IMPORT	|Image$$ZI$$Base|
	IMPORT	|Image$$ZI$$Limit|

	ENTRY
|__main|
VectorsStart
	; The hardware vector code
	LDR	PC, ResetVectorAddr
	LDR	PC, UndefinedVectorAddr
	LDR	PC, SWIVectorAddr
	LDR	PC, PrefetchAbortVectorAddr
	LDR	PC, DataAbortVectorAddr
	NOP							; not used
	LDR	PC, IRQVectorAddr
	LDR	PC, FIQVectorAddr

ResetVectorAddr
	DCD	ResetVector
UndefinedVectorAddr
	DCD	ExceptH_EntryUndefV
SWIVectorAddr
	DCD	ExceptH_EntrySWIV
PrefetchAbortVectorAddr
	DCD	ExceptH_EntryPAbortV
DataAbortVectorAddr
	DCD	ExceptH_EntryDAbortV
; not used
	NOP
IRQVectorAddr
	DCD	ExceptH_EntryIRQV
FIQVectorAddr
	DCD	ExceptH_EntryFIQV
VectorsEnd

	LTORG

ResetVector
	; Right, the CPU enters here in SVC?, IRQs & FIQs off

	; Under Demon or Angel, make sure we're in SVC
   IF :DEF: DEMON	
	SWI	OS_EnterOS		; This is SVC mode code.
   ENDIF

   IF :DEF: ANGEL
	MOV	R0, #angel_SWIreason_EnterSVC
	SWI	0x123456		; Returns EnterUSR routine in r0
   ENDIF

	; Get us into a state we know (various ARM's differ in start up mode)
	MRS	R0, CPSR
	BIC	R0, R0, #0x1f
	ORR	R0, R0, #IBit :OR: FBit :OR: Mode_SVC32
	MSR	CPSR_all, R0

	[ :DEF: ANGEL
	|
	; Remap ROM
	BOARD_MOVEROM

	; Get DRAM running
	BOARD_INITIALISEDRAM
	]

	; Find out where our zero page lives
	; Returns our zero page memory
	BOARD_RETURNZPLOCATION R9

	; Copy our processor vectors into the zero page
	ADR	R1,VectorsStart
	ADR	R2,VectorsEnd
	MOV	R3,R9
1	LDR	R0,[R1],#4
	STR	R0,[R3],#4
	CMP	R1,R2
	BLO	%b01

	; Setup processor stacks
	ADD	R3,R9,#BOARD_HALSTKSLOCATION
	MRS	R2,CPSR
	BIC	R4,R2,#0x1f

	; Setup IRQ32 stack
	ORR	R2,R4,#Mode_IRQ32
	MSR	CPSR_all,R2						; Now in IRQ32
	NOP
	MOV	R13,R3
	SUB	R3,R3,#IRQ32StackSize

	; Setup FIQ32 stack
	ORR	R2,R4,#Mode_FIQ32
	MSR	CPSR_all,R2						; Now in FIQ32
	NOP
	MOV	R13,R3
	SUB	R3,R3,#FIQ32StackSize

	; Setup Abort32 stack
	ORR	R2,R4,#Mode_Abort32
	MSR	CPSR_all,R2						; Now in Abort32
	NOP
	MOV	R13,R3
	SUB	R3,R3,#Abort32StackSize

	; Setup Undef32 stack
	ORR	R2,R4,#Mode_Undef32
	MSR	CPSR_all,R2						; Now in Undef32
	NOP
	MOV	R13,R3
	SUB	R3,R3,#Undef32StackSize

	; Setup SVC32 stack
	ORR	R2,R4,#Mode_SVC32
	MSR	CPSR_all,R2						; Now in SVC32
	NOP
	MOV	R13,R3
	SUB	R3,R3,#SVC32StackSize

	; Proceed with ARM initialisation	
	BL	ARMCore_Initialise				; Determine our CPU
	BL	ARMMMU_Initialise				; Determine our MMU
	BL	ARMCache_Initialise				; Determine our Cache

	; Enable 32 bit program and data spaces
	MOV	R0,#ARMCORE_SETP32 :OR: ARMCORE_SETD32
	MOV	R1,#ARMCORE_SETP32 :OR: ARMCORE_SETD32
	BL	ARMCore_Set

	; Determine memory configuration, set up MMU tables and enable caches
	BOARD_SETUPMEMORYSYSTEM

	; Setup User32 stack at top of memory
	LDR	R1,[R9,#HALZP_DRAMTOP]
	STMDB	R13,{R1}
	LDMDB	R13,{R13}^					; Sets r13_user32

	; Copy C variables into RAM
	LDR	R4,=|Image$$RO$$Limit|
	LDR	R5,=|Image$$RW$$Base|
	LDR	R6,=|Image$$RW$$Limit|
1
	CMP	R5,R6
	BEQ	%f02
	LDR	R2,[R4],#4
	STR	R2,[R5],#4
	B	%b01
2
	MOV	R2,#0
	LDR	R5,=|Image$$ZI$$Base|
	LDR	R6,=|Image$$ZI$$Limit|
3
	CMP	R5,R6
	BEQ	%f04
	STR	R2,[R5],#4
	B	%b03
4

	; We're now able to call C code
	; Set up sl just in case it's needed
	MOV	R10,R13

	; Initialise DbgIO module
	BL	DbgIO_Initialise

	[ :DEF: ANGEL
	; Point DbgIO at the Angel UART driver
	ADR	R0,DbgIO_SetBlk
	ADD	R1,R9,#BOARD_HALZP_UARTSPACE
	ADD	R1,R1,#HALZP_ANGELUARTSPACE
	MOV	R2,#0						; not relevent here
	BL	DbgIO_Set
	]

	[ :DEF: DEMON
	; Point DbgIO at the Demon UART driver
	ADR	R0,DbgIO_SetBlk
	ADD	R1,R9,#BOARD_HALZP_UARTSPACE
	ADD	R1,R1,#HALZP_DEMONUARTSPACE
	MOV	R2,#0						; not relevent here
	BL	DbgIO_Set
	]

	; Initialise Exceptions Handler module
	BL	ExceptH_Initialise

	; Set a SWI handler to something default (v. important as many calls
	; perform a OS_EnterOS if called in User32)
	MOV	R0,#SWIV
	ADR	R1,SWIEHandler
	MOV	R2,#0
	MOV	R3,#2						; use R12
	BL	ExceptH_AddHandler
	CMP	R0,#0
	BNE	NedHAL_EmergencyStop

	[ :DEF: STANDALONE
	|
	; Install the ExceptH hooks over the current hardware vectors
	MOV	R4,#0
	MOV	R5,#1
1
	LDR	R1,[R4,R5,LSL #2]				; load the LDR PC instruction
	MOV	R0,#8						; offset for pipeline
	ADD	R0,R0,R5,LSL #2
	MOV	R1,R1,LSL #20					; mask out offset
	ADD	R0,R0,R1,LSR #20				; add offset to PC of that instr
	LDR	R1,[R0]						; load actual address of routine
	
	MOV	R0,R5
	MOV	R2,#0
	MOV	R3,#4						; former function
	CMP	R0,#5						; skip vector no 5 (no longer used)
	MOVEQ	R0,#0
	BLNE	ExceptH_AddHandler
	CMP	R0,#0
	BNE	NedHAL_EmergencyStop
	ADD	R5,R5,#1
	CMP	R5,#7
	BLE	%b01

	ADR	R1,VectorsStart
	ADD	R1,R1,#4
	ADR	R2,VectorsEnd
	MOV	R3,#4

	LDR	R0,[R1,#8-4]					; only SWI for now
	STR	R0,[R3,#8-4]
	LDR	R0,[R1,#40-4]
	STR	R0,[R3,#40-4]

	LDR	R0,[R1,#24-4]					; only IRQ for now
	STR	R0,[R3,#24-4]
	LDR	R0,[R1,#56-4]
	STR	R0,[R3,#56-4]

;1	LDR	R0,[R1],#4
;	STR	R0,[R3],#4
;	CMP	R1,R2
;	BLO	%b01
	]

	; Fall into USR mode with interrupts enabled
	MRS	R2,CPSR
	BIC	R4,R2,#0x1f
	ORR	R2,R4,#Mode_User32
	BIC	R2,R2,#IBit :OR: FBit
	MSR	CPSR_all,R2						; Now in User32
	NOP

	; Do any board specific initialisations
	BOARD_INITIALISE

	; Call OS
	BL	NedHAL_Entry

	; Do any board specific finalisations
	BOARD_FINALISE

	; Kill Exceptions Handler module
	; BL	ExceptH_Finalise

	; Print default stop message
	ADR	R0,NedHAL_Entry_StopErrBlk

NedHAL_EmergencyStop
	; R0 contains error blk
	; module id is at HALERROR_MODULE (8 bytes, non terminated)
	; module code is at HALERROR_CODE (4 bytes)
	; error msg ptr is at HALERROR_NESSAGE (4 bytes)
	MOV	R9,R0
	LDMIA	R0,{R1,R2}
	SUB	R13,R13,#12
	STMIA	R13,{R1,R2}
	MOV	R0,#0
	STR	R0,[R13,#8]
	ADR	R0,NedHAL_EmergencyStopTxt1
	MOV	R1,R13
	LDR	R2,[R9,#HALERROR_CODE]
	BL	DbgIO_WriteF
	LDR	R0,[R9,#HALERROR_MESSAGE]
	BL	DbgIO_Write0
	ADR	R0,StopTxt
	BL	DbgIO_Write0
	; Kill DbgIO
	BL	DbgIO_Finalise

	; Terminate if debugger is present
   IF :DEF: DEMON
	SWI	OS_Exit
   ENDIF

   IF :DEF: ANGEL
	LDR	r0, =angel_SWIreason_ReportException
	LDR	r1, =ADP_Stopped_ApplicationExit
	SWI	0x123456
   ENDIF
	; If no debugger, loop forever
	B	.
NedHAL_EmergencyStopTxt1
	DCB	"EMERGENCY HALT IN MODULE %s (module code %d):",10,13,0
	ALIGN
StopTxt	DCB	10,13,"STOP!",10,13,0
	ALIGN
NedHAL_Entry_StopErrBlk
	DCD	0
	DCB	"NedHAL_Entry() returned, stopping system!",0
	ALIGN


	LTORG

	[ :DEF: DEMON
	; Keep the same as the DbgIO_UARTDetails structure in Core/io/stdout.h
	IMPORT	DemonUARTDriver_Initialise
	IMPORT	DemonUARTDriver_Finalise
	IMPORT	DemonUARTDriver_Send
	IMPORT	DemonUARTDriver_Get

DbgIO_SetBlk
	DCD	DemonUARTDriver_Initialise
	DCD	DemonUARTDriver_Finalise
	DCD	DemonUARTDriver_Send
	DCD	DemonUARTDriver_Get
	]

	[ :DEF: ANGEL
	; Keep the same as the DbgIO_UARTDetails structure in Core/io/stdout.h
	IMPORT	AngelUARTDriver_Initialise
	IMPORT	AngelUARTDriver_Finalise
	IMPORT	AngelUARTDriver_Send
	IMPORT	AngelUARTDriver_Get

DbgIO_SetBlk
	DCD	AngelUARTDriver_Initialise
	DCD	AngelUARTDriver_Finalise
	DCD	AngelUARTDriver_Send
	DCD	AngelUARTDriver_Get
	]

|_main|
	MOV	PC,R14

SWIEHandler
	; Very basic support for three handy processor mode change SWIs
	; We have been called by ExceptH, hence we have R0-R3 available
	; SPSR, R14, R0-R3 are on the stack
	LDR	R0,[R13,#4]					; addr of SWI + 4
	LDR	R0,[R0,#-4]
	BIC	R0,R0,#0xff000000
	CMP	R0,#OS_IntOn
	BEQ	SWIVector_IntOn
	CMP	R0,#OS_IntOff
	BEQ	SWIVector_IntOff
	CMP	R0,#OS_EnterOS
	BEQ	SWIVector_EnterOS
	; An unknown. Pass on the call
	MOV	R0,#0
	MOV	PC,R14
SWIVector_IntOn
	; Just enable interrupts, leaving everything else alone
	LDR	R0,[R13,#0]
	BIC	R0,R0,#IBit
	STR	R0,[R13,#0]
	MOV	R0,#1						; claimed
	MOV	PC,R14
SWIVector_IntOff
	; Just disable interrupts, leaving everything else alone
	LDR	R0,[R13,#0]
	ORR	R0,R0,#IBit
	STR	R0,[R13,#0]
	MOV	R0,#1						; claimed
	MOV	PC,R14
SWIVector_EnterOS
	; Just enter SVC on exit, leaving everything else alone
	LDR	R0,[R13,#0]
	BIC	R0,R0,#Mode_Bits
	ORR	R0,R0,#Mode_SVC32
	STR	R0,[R13,#0]
	MOV	R0,#1						; claimed
	MOV	PC,R14


	LTORG

	END
