; Core/ARM/Excepts.s
; Exception handler management module
;
; 19/Oct/1999 ned

	; Include common stuff
	INCLUDE Core/ARM/CommonH.s

	AREA Exceptionhandling, CODE, READONLY

	EXPORT	ExceptH_Initialise
	EXPORT	ExceptH_AddHandler
	EXPORT	ExceptH_RemoveHandler
	EXPORT	ExceptH_Finalise

	EXPORT	ExceptH_EntryUndefV
	EXPORT	ExceptH_EntrySWIV
	EXPORT	ExceptH_EntryPAbortV
	EXPORT	ExceptH_EntryDAbortV
	EXPORT	ExceptH_EntryIRQV
	EXPORT	ExceptH_EntryFIQV

	IMPORT	NedHAL_EmergencyStop

ExceptH_Initialise
	STMFD R13!,{R9,R10,R14}
	BOARD_RETURNZPLOCATION	R9
	; Ensure privilaged mode
	MRS	R10,CPSR
	TST	R10,#IBit
	BNE	ExceptH_Initialise_IOff
	AND	R3,R10,#Mode_Bits
	CMP	R3,#Mode_User32
	SWINE	OS_EnterOS
	MRS	R3,CPSR
	ORR	R3,R3,#IBit
	MSR	CPSR_all,R4
ExceptH_Initialise_IOff
	; Zero the handler lists
	MOV	R1,#0
	MOV	R2,#0
	ADD	R3,R9,#HALZP_END_OF_HANDLERS
	ADD	R0,R9,#HALZP_START_OF_HANDLERS
1
	STMDB	R3,{R1,R2,R3}					; *** NOTE: ASSUMES SIZE OF STRUCTURE
	SUB	R3,R3,#3*4
	CMP	R3,R0
	BHS	%b01

	MOV	R0,#0
	LDMFD	R13!,{R9,R10,PC}

ExceptH_AddHandler
	; R0 is vector number
	; R1 is addr of handler
	; R2 is value to pass to handler
	; R3 is flags (0: put handler at top, 1: pass R2 in R12, 2: restore context before call)
	STMFD	R13!,{R4-R6,R8,R9,R10,R14}
	BOARD_RETURNZPLOCATION	R9
	; Decide which vector we're on about
	CMP	R0,#ResetV
	MOVEQ	R9,#0
	MOVEQ	R8,#0
	CMP	R0,#UndefV
	ADDEQ	R9,R9,#HALZP_UNDEFE_HANDLERS
	MOVEQ	R8,#UndefinedExcptHandlers
	CMP	R0,#SWIV
	ADDEQ	R9,R9,#HALZP_SWIE_HANDLERS
	MOVEQ	R8,#SWIExcptHandlers
	CMP	R0,#PAbortV
	ADDEQ	R9,R9,#HALZP_PABORTE_HANDLERS
	MOVEQ	R8,#PAbortExcptHandlers
	CMP	R0,#DAbortV
	ADDEQ	R9,R9,#HALZP_DABORTE_HANDLERS
	MOVEQ	R8,#DAbortExcptHandlers
	CMP	R0,#IRQV
	ADDEQ	R9,R9,#HALZP_IRQE_HANDLERS
	MOVEQ	R8,#IRQExcptHandlers
	CMP	R0,#FIQV
	ADDEQ	R9,R9,#HALZP_FIQE_HANDLERS
	MOVEQ	R8,#FIQExcptHandlers
	
	; Ensure we have some handlers
	CMP	R8,#0
	ADREQ	R0,ExceptH_AddHandler_NoSlotsErrBlk
	LDMEQFD	R13!,{R4-R6,R8,R9,R10,PC}

	; Ensure privilaged mode
	MRS	R10,CPSR
	TST	R10,#IBit
	BNE	ExceptH_AddHandler_IOff
	AND	R4,R10,#Mode_Bits
	CMP	R4,#Mode_User32
	SWINE	OS_EnterOS
	MRS	R4,CPSR
	ORR	R4,R4,#IBit
	MSR	CPSR_all,R4
ExceptH_AddHandler_IOff
	; R0=vector no, R1=handler addr, R2=value, R3=flags
	; R4-R6=scratch, R8=no of handlers, R9=handlers list base
	; R10=old CPSR
	[ :DEF: STANDALONE
	TST	R3,#1
	BEQ	ExceptH_AddHandler_InsertAtZero
	|
	TST	R3,#4						; insert all normals at start
	BNE	ExceptH_AddHandler_InsertAtZero			; only the former at end (zero)
	]
ExceptH_AddHandler_InsertAtStart
	MOV	R5,#HALZP_HANDLERS_SIZEOF
	MLA	R4,R8,R5,R9					; *HALZP_HANDLERS_SIZEOF
	LDR	R5,[R3,#-HALZP_HANDLERS_SIZEOF]			; get last entry
	CMP	R5,#0
	BNE	ExceptH_AddHandler_NoSpaces
	; Make space by moving everything down HALZP_HANDLERS_SIZEOF bytes
1
	SUB	R4,R4,#2*HALZP_HANDLERS_SIZEOF			; point at penultimate
	LDMIA	R4!,{R0,R5,R6}					; *** NOTE: ASSUMES SIZE OF STRUCTURE
	STMIA	R4,{R0,R5,R6}
	SUB	R4,R4,#HALZP_HANDLERS_SIZEOF
	CMP	R4,R9
	BHI	%b01
	ADD	R9,R9,#HALZP_HANDLERS_SIZEOF			; point at first entry
	B	ExceptH_AddHandler_Install
ExceptH_AddHandler_InsertAtZero
	MOV	R4,#0
1
	LDR	R5,[R9],#HALZP_HANDLERS_SIZEOF
	CMP	R5,#0
	BEQ	ExceptH_AddHandler_Install
	ADD	R4,R4,#1
	CMP	R4,R8
	BLT	%b01
	; Falls through into ExceptH_AddHandler_NoSpaces
ExceptH_AddHandler_NoSpaces
	; No spaces, better return an error
	ADR	R0,ExceptH_AddHandler_NoSlotsErrBlk
	MSR	CPSR_all,R10
	NOP
	LDMFD	R13!,{R4-R6,R8,R9,R10,PC}
ExceptH_AddHandler_Install
	STMDB	R9,{R1,R2,R3}					; *** NOTE: ASSUMES SIZE OF STRUCTURE
	MOV	R0,#0
	MSR	CPSR_all,R10
	NOP
	LDMFD	R13!,{R4-R6,R8,R9,R10,PC}
ExceptH_AddHandler_NoSlotsErrBlk
	DCB	"ExceptH",0
	DCD	2
	DCD	ExceptH_AddHandler_NoSlotsErrBlkMsg
ExceptH_AddHandler_NoSlotsErrBlkMsg
	DCB	"All handler slots filled. Increase appropriate handler no if necessary",0
	ALIGN

ExceptH_RemoveHandler
	; R0 is vector no
	; R1 is addr of handler
	STMFD	R13!,{R4,R8,R9,R10,R14}
	BOARD_RETURNZPLOCATION	R9
	; Decide which vector we're on about
	CMP	R0,#ResetV
	MOVEQ	R9,#0
	MOVEQ	R8,#0
	CMP	R0,#UndefV
	ADDEQ	R9,R9,#HALZP_UNDEFE_HANDLERS
	MOVEQ	R8,#UndefinedExcptHandlers
	CMP	R0,#SWIV
	ADDEQ	R9,R9,#HALZP_SWIE_HANDLERS
	MOVEQ	R8,#SWIExcptHandlers
	CMP	R0,#PAbortV
	ADDEQ	R9,R9,#HALZP_PABORTE_HANDLERS
	MOVEQ	R8,#PAbortExcptHandlers
	CMP	R0,#DAbortV
	ADDEQ	R9,R9,#HALZP_DABORTE_HANDLERS
	MOVEQ	R8,#DAbortExcptHandlers
	CMP	R0,#IRQV
	ADDEQ	R9,R9,#HALZP_IRQE_HANDLERS
	MOVEQ	R8,#IRQExcptHandlers
	CMP	R0,#FIQV
	ADDEQ	R9,R9,#HALZP_FIQE_HANDLERS
	MOVEQ	R8,#FIQExcptHandlers
	
	; Ensure we have some handlers
	CMP	R8,#0
	ADREQ	R0,ExceptH_RemoveHandler_NotPresentErrBlk
	LDMEQFD	R13!,{R4,R8,R9,R10,PC}

	; Ensure privilaged mode
	MRS	R10,CPSR
	TST	R10,#IBit
	BNE	ExceptH_RemoveHandler_IOff
	AND	R4,R10,#Mode_Bits
	CMP	R4,#Mode_User32
	SWINE	OS_EnterOS
	MRS	R4,CPSR
	ORR	R4,R4,#IBit
	MSR	CPSR_all,R4
ExceptH_RemoveHandler_IOff

	MOV	R4,#0
1
	LDR	R0,[R9],#HALZP_HANDLERS_SIZEOF
	CMP	R0,R1
	BEQ	%f02
	ADD	R4,R4,#1
	CMP	R4,R8
	BLT	%b01
	ADR	R0,ExceptH_RemoveHandler_NotPresentErrBlk
	LDMFD	R13!,{R4,R8,R9,R10,PC}

2
	; Copy everything up HALZP_HANDLERS_SIZEOF bytes
	LDMIA	R9,{R0,R1,R2}					; *** NOTE: ASSUMES SIZE OF STRUCTURE
	STMDB	R9,{R0,R1,R2}
	ADD	R9,R9,#HALZP_HANDLERS_SIZEOF
	ADD	R4,R4,#1
	CMP	R4,R8
	BLT	%b02
	MOV	R0,#0
	MOV	R1,#0
	MOV	R2,#0
	STMDB	R9,{R0,R1,R2}
	LDMFD	R13!,{R4,R8,R9,R10,PC}
ExceptH_RemoveHandler_NotPresentErrBlk
	DCB	"ExceptH",0
	DCD	3
	DCD	ExceptH_RemoveHandler_NotPresentErrBlkMsg
ExceptH_RemoveHandler_NotPresentErrBlkMsg
	DCB	"Unable to find IRQ Handler specified",0
	ALIGN

ExceptH_Finalise
	B	ExceptH_Initialise

ExceptH_UnwindAndCall
	; R1=handler addr
	LDMFD	R13!,{R0,R14}
	MSR	SPSR_all, R0					; restore SPSR

	ADD	R8,R13,#9*4					; r13 when we entered ExceptH_Entry + SPSR
	MOV	R12,R13
	MOV	R9,R8
	SUB	R13,R13,#11*4					; create a restoration stack frame
	STMDB	R12!,{R1}					; PC
	STMDB	R12!,{R8}					; old R13
	LDMDB	R9!,{R0-R3}					; old R9,R10,R11,R12
	STMDB	R12!,{R0-R3}
	LDMDB	R9!,{R0-R3,R8}					; old R0-R3,R8
	STMDB	R12!,{R0-R3,R8}
	LDMFD	R12,{R0-R3,R8-R12,R13,PC}			; restores all regs and calls old routine

ExceptH_EntryUndefV
	B	.

ExceptH_EntrySWIV
	; ARM enters here when the ARM SWI vector is branched to
	; R14_svc contains return addr, SPSR_svc contains saved state
	STMFD	R13!,{R0-R3,R8-R12}
	MRS	R0,SPSR_all
	STMFD	R13!,{R0,R14}
	BOARD_RETURNZPLOCATION	R9
	ADD	R9,R9,#HALZP_SWIE_HANDLERS
	ADD	R8,R9,#SWIExcptHandlers * HALZP_HANDLERS_SIZEOF
1
	LDMIA	R9!,{R1,R2,R3}					; *** NOTE: ASSUMES SIZE OF STRUCTURE
	CMP	R9,R8
	BHI	ExceptH_EntrySWIV_NotClaimed
	CMP	R1,#0
	BEQ	ExceptH_EntrySWIV_NotClaimed
	TST	R3,#4						; handler is unaware of us
	ADRNE	R14,%b01
	BNE	ExceptH_UnwindAndCall

	MOV	R10,R2
	MOV	R11,R1
	ADD	R12,R13,#8					; skip SPSR, R14
	TST	R3,#2						; Use R12 not R0
	LDMFD	R12,{R0-R3}					; Preserve R0-R7 on entry to handler
	MOVNE	R12,R10
	MOVEQ	R0,R10
	SUB	R10,R13,#SVC32StackSize				; set sl for stk checking code
	MOV	R14,PC
	MOV	PC,R11
	; returns here
	CMP	R0,#0
	BEQ	%b01
	CMP	R0,#1						; claimed?
	BEQ	ExceptH_EntrySWIV_Exit
	; must be an error blk then - oh dear!
	BL	NedHAL_EmergencyStop
	B	ExceptH_EntrySWIV_Exit
ExceptH_EntrySWIV_NotClaimed
	; Okay, nothing's claimed it - take default action
	ADR	R0,ExceptH_EntrySWIV_NotClaimedErrBlk
	BL	NedHAL_EmergencyStop
ExceptH_EntrySWIV_Exit
	LDMFD	R13!,{R0,R14}
	MSR	SPSR_all,R0
	LDMFD	R13!,{R0-R3,R8-R12}
	MOVS	PC,R14
ExceptH_EntrySWIV_NotClaimedErrBlk
	DCB	"ExceptH",0
	DCD	128+1
	DCD	ExceptH_EntrySWIV_NotClaimedErrBlkMsg
ExceptH_EntrySWIV_NotClaimedErrBlkMsg
	DCB	"SWI Exception unclaimed",0
	ALIGN

ExceptH_EntryPAbortV
	B	.

ExceptH_EntryDAbortV
	B	.

ExceptH_EntryIRQV
	; ARM enters here when the ARM IRQ vector is branched to
	; R14_irq contains return addr, SPSR_irq contains saved state
	STMFD	R13!,{R0-R3,R8-R12}
	MRS	R0,SPSR_all
	STMFD	R13!,{R0,R14}
	BOARD_RETURNZPLOCATION	R9
	ADD	R9,R9,#HALZP_IRQE_HANDLERS
	ADD	R8,R9,#IRQExcptHandlers * HALZP_HANDLERS_SIZEOF
1
	LDMIA	R9!,{R1,R2,R3}					; *** NOTE: ASSUMES SIZE OF STRUCTURE
	CMP	R9,R8
	BHI	ExceptH_EntryIRQV_NotClaimed
	CMP	R1,#0
	BEQ	ExceptH_EntryIRQV_NotClaimed
	TST	R3,#4						; handler is unaware of us
	ADRNE	R14,%b01
	BNE	ExceptH_UnwindAndCall

	MOV	R10,R2
	MOV	R11,R1
	ADD	R12,R13,#8					; skip SPSR, R14
	TST	R3,#2						; Use R12 not R0
	LDMFD	R12,{R0-R3}					; Preserve R0-R7 on entry to handler
	MOVNE	R12,R10
	MOVEQ	R0,R10
	SUB	R10,R13,#SVC32StackSize				; set sl for stk checking code
	MOV	R14,PC
	MOV	PC,R11
	; returns here
	CMP	R0,#0
	BEQ	%b01
	CMP	R0,#1						; claimed?
	BEQ	ExceptH_EntryIRQV_Exit
	; must be an error blk then - oh dear!
	BL	NedHAL_EmergencyStop
	B	ExceptH_EntryIRQV_Exit
ExceptH_EntryIRQV_NotClaimed
	; Okay, nothing's claimed it - take default action
	ADR	R0,ExceptH_EntryIRQV_NotClaimedErrBlk
	BL	NedHAL_EmergencyStop
ExceptH_EntryIRQV_Exit
	LDMFD	R13!,{R0,R14}
	MSR	SPSR_all,R0
	LDMFD	R13!,{R0-R3,R8-R12}
	MOVS	PC,R14
ExceptH_EntryIRQV_NotClaimedErrBlk
	DCB	"ExceptH",0
	DCD	128+5
	DCD	ExceptH_EntryIRQV_NotClaimedErrBlkMsg
ExceptH_EntryIRQV_NotClaimedErrBlkMsg
	DCB	"IRQ Exception unclaimed",0
	ALIGN


ExceptH_EntryFIQV
	B	.


	END
