/* Core/io/fancyprint.c
Implements a useful part-ANSI C printf style function

14/Oct/1999 ned
*/

#include "DebugIO.h"

/* NOTE: Requires the following embedded ANSI C functions */
#include <ctype.h>
#include <stdarg.h>

#ifdef __cplusplus
extern "C" {
#endif

static int vsprintf(char *buff, const char *format, va_list aplist)
{
	const char *fptr=format;
	char *buffptr=buff;

	while(*fptr)
	{
		if(*fptr=='%')
		{
			switch(*(++fptr))
			{
				case '%':
				{
					*(buffptr++)='%';
					break;
				}
				case 'd':
				{
					/* Take a number and print it */
					int div=1, no;
					no=va_arg(aplist, int);
					if(no<0)
					{
						*(buffptr++)='-';
						no=-no;
					}
					while(no/div>=10) div*=10;
					do
					{
						*(buffptr++)='0'+((no / div) % 10);
						div/=10;
					} while(div);
					break;
				}
				case 'x':
				{
					/* Take a number and print it in hex*/
					int div=1, no;
					no=va_arg(aplist, int);
					while(no/div>=16) div*=16;
					do
					{
						int t=(no / div) & 0xf;
						if(t>=10)
						{
							*(buffptr++)='A'+t-10;
						}
						else
						{
							*(buffptr++)='0'+t;
						}
						div/=16;
					} while(div);
					break;
				}
				case 's':
				{
					/* Take a string and print it */
					char *str;
					str=va_arg(aplist, char *);
					while(*str) *(buffptr++)=*(str++);
					break;
				}
				default:
				{
					/* Print a %? */
					*(buffptr++)='%';
					*(buffptr++)='?';
					break;
				}
			}
			fptr++;
		}
		else *(buffptr++)=*(fptr++);
	}
	*buffptr=0;
	return (buffptr-buff);
}

extern HALError *DbgIO_WriteF(const char *format, ...)
{
	char buff[256];
	va_list vargs;
	HALError *ret;

	va_start(vargs, format);
	vsprintf(buff, format, vargs);
	va_end(vargs);

	ret=DbgIO_Write0(buff);
	return ret;
}

#ifdef __cplusplus
}
#endif

